﻿local cfg = oUF_Hank_config

-- ButtonFacade support
local LBF, BFAuras, BFHighlight, BFDummy
if IsAddOnLoaded("ButtonFacade") then
	LBF = LibStub("LibButtonFacade", true)
	BFAuras = LBF:Group("oUF_Hank", "Auras")
	BFHighlight = LBF:Group("oUF_Hank", "Aura highlight")
	BFDummy = CreateFrame("Frame")
	BFDummy:Hide()
end

local digitTexCoords = {
	["1"] = {1, 20},
	["2"] = {21, 31},
	["3"] = {53, 30},
	["4"] = {84, 33},
	["5"] = {118, 30},
	["6"] = {149, 31},
	["7"] = {181, 30},
	["8"] = {212, 31},
	["9"] = {244, 31},
	["0"] = {276, 31},
	["%"] = {308, 17},
	["X"] = {326, 31}, -- Dead
	["G"] = {358, 36}, -- Ghost
	["Off"] = {395, 23}, -- Offline
	["height"] = 42,
	["texWidth"] = 512,
	["texHeight"] = 128
}

local fntBig = CreateFont("UFFontBig")
fntBig:SetFont(unpack(cfg.FontStyleBig))
local fntMedium = CreateFont("UFFontMedium")
fntMedium:SetFont(unpack(cfg.FontStyleMedium))
fntMedium:SetTextColor(unpack(cfg.colors.text))
fntMedium:SetShadowColor(unpack(cfg.colors.textShadow))
fntMedium:SetShadowOffset(1, -1)
local fntSmall = CreateFont("UFFontSmall")
fntSmall:SetFont(unpack(cfg.FontStyleSmall))
fntSmall:SetTextColor(unpack(cfg.colors.text))
fntSmall:SetShadowColor(unpack(cfg.colors.textShadow))
fntSmall:SetShadowOffset(1, -1)

local canDispel = {
	["DRUID"] = {["Poison"] = true, ["Curse"] = true},
	["MAGE"] = {["Curse"] = true},
	["PALADIN"] = {["Disease"] = true, ["Poison"] = true, ["Magic"] = true},
	["PRIEST"] = {["Magic"] = true, ["Disease"] = true},
	["SHAMAN"] = {["Poison"] = true, ["Disease"] = true, ["Curse"] = ({GetTalentInfo(3, 18, false, false, nil)})[5] == 1}, -- Curse only with Cleanse Spirit talented
	["WARLOCK"] = {["Magic"] = true},
	["HUNTER"] = {},
	["ROGUE"] = {},
	["WARRIOR"] = {},
	["DEATHKNIGHT"] = {}
}

-- Functions -------------------------------------

-- Unit menu
local function menu(self)
	local unit = self.unit:sub(1, -2)
	local cunit = self.unit:gsub("(.)", string.upper, 1)

	if(unit == "party" or unit == "partypet") then
		ToggleDropDownMenu(1, nil, _G["PartyMemberFrame"..self.id.."DropDown"], "cursor", 0, 0)
	elseif(_G[cunit.."FrameDropDown"]) then
		ToggleDropDownMenu(1, nil, _G[cunit.."FrameDropDown"], "cursor", 0, 0)
	end
end

-- Shorten health and power numbers
local function valShort(value)
	if(value >= 1e6) then
		return ("%.2f"):format(value / 1e6):gsub("%.?0+$", "") .. "m"
	elseif(value >= 1e4) then
		return ("%.1f"):format(value / 1e3):gsub("%.?0+$", "") .. "k"
	else
		return value
	end
end

-- Party frame be gone!
local function HideParty()
	for i = 1, 4 do
		local party = "PartyMemberFrame" .. i
		local frame = _G[party]

		frame:UnregisterAllEvents()
		frame.Show = function() end
		frame:Hide()

		_G[party .. "HealthBar"]:UnregisterAllEvents()
		_G[party .. "ManaBar"]:UnregisterAllEvents()
	end
end

-- This is where the magic happens. Handle health update, display digit textures
local function UpdateHealth(self, unit)
	local h, hMax

	-- In vehicle
	if unit == "player" and UnitHasVehicleUI("player") then
		h, hMax = UnitHealth("pet"), UnitHealthMax("pet")
	else
		h, hMax = UnitHealth(unit), UnitHealthMax(unit)
	end

	local status = (not UnitIsConnected(unit) or nil) and "Off" or UnitIsGhost(unit) and "G" or UnitIsDead(unit) and "X"

	if not status then
		local hPerc = ("%d%%"):format(h / hMax * 100 + 0.5)
		local len = string.len(hPerc)

		for i = 1, 4 do
			if i > len then
				self.health[5 - i]:Hide()
				self.healthFill[5 - i]:Hide()
			else
				local digit
				if self == oUF_player then
					digit = string.sub(hPerc , -i, -i)
				elseif self == oUF_target or self == oUF_focus then
					digit = string.sub(hPerc , i, i)
				end
				self.health[5 - i]:SetSize(digitTexCoords[digit][2], digitTexCoords["height"])
				self.health[5 - i]:SetTexCoord(digitTexCoords[digit][1] / digitTexCoords["texWidth"], (digitTexCoords[digit][1] + digitTexCoords[digit][2]) / digitTexCoords["texWidth"], 1 / digitTexCoords["texHeight"], (1 + digitTexCoords["height"]) / digitTexCoords["texHeight"])
				self.health[5 - i]:Show()
				self.healthFill[5 - i]:SetSize(digitTexCoords[digit][2], digitTexCoords["height"] * h / hMax)
				self.healthFill[5 - i]:SetTexCoord(digitTexCoords[digit][1] / digitTexCoords["texWidth"], (digitTexCoords[digit][1] + digitTexCoords[digit][2]) / digitTexCoords["texWidth"], (2 + 2 * digitTexCoords["height"] - digitTexCoords["height"] * h / hMax) / digitTexCoords["texHeight"], (2 + 2 * digitTexCoords["height"]) / digitTexCoords["texHeight"])
				self.healthFill[5 - i]:Show()
			end
		end

		if self == oUF_player then
			self.power:SetPoint("BOTTOMRIGHT", self.health[5 - len], "BOTTOMLEFT", -5, 0)
		elseif self == oUF_target or self == oUF_focus then
			self.power:SetPoint("BOTTOMLEFT", self.health[5 - len], "BOTTOMRIGHT", 5, 0)
		end
	else
		for i = 1, 4 do
			self.healthFill[i]:Hide()
			self.health[i]:Hide()
		end

		self.health[4]:SetSize(digitTexCoords[status][2], digitTexCoords["height"])
		self.health[4]:SetTexCoord(digitTexCoords[status][1] / digitTexCoords["texWidth"], (digitTexCoords[status][1] + digitTexCoords[status][2]) / digitTexCoords["texWidth"], 1 / digitTexCoords["texHeight"], (1 + digitTexCoords["height"]) / digitTexCoords["texHeight"])
		self.health[4]:Show()

		if self == oUF_player then
			self.power:SetPoint("BOTTOMRIGHT", self.health[4], "BOTTOMLEFT", -5, 0)
		elseif self == oUF_target or self == oUF_focus then
			self.power:SetPoint("BOTTOMLEFT", self.health[4], "BOTTOMRIGHT", 5, 0)
		end
	end
end

-- Reanchor icons on player status change
local function PostUpdateStatus(self)
	-- Icon order: (Combat) (Rested) (Leader)
	local leader = IsPartyLeader()
	local resting = IsResting()
	local incombat = UnitAffectingCombat("player")

	if leader then
		self.Leader:ClearAllPoints()
		self.Leader:SetPoint("BOTTOMRIGHT", self.power, "TOPRIGHT")
		if resting then
			self.Resting:ClearAllPoints()
			self.Resting:SetPoint("RIGHT", self.Leader, "LEFT")
			if incombat then
				self.Combat:ClearAllPoints()
				self.Combat:SetPoint("RIGHT", self.Resting, "LEFT")
			end
		elseif incombat then
			self.Combat:ClearAllPoints()
			self.Combat:SetPoint("RIGHT", self.Leader, "LEFT")
		end
	elseif resting then
		self.Resting:ClearAllPoints()
		self.Resting:SetPoint("BOTTOMRIGHT", self.power, "TOPRIGHT")
		if incombat then
			self.Combat:ClearAllPoints()
			self.Combat:SetPoint("RIGHT", self.Resting, "LEFT")
		end
	elseif incombat then
		self.Combat:ClearAllPoints()
		self.Combat:SetPoint("BOTTOMRIGHT", self.power, "TOPRIGHT")
	end
end

-- Reanchoring / -sizing on name update
local function PostUpdateName(unit, self)
	if unit == "target" or unit == "focus" then
		-- Reanchor raid icon to the largest string (either name or power)
		if self.name:GetWidth() >= self.power:GetWidth() then
			self.RaidIcon:SetPoint("LEFT", self.name, "RIGHT", 10, 0)
		else
			self.RaidIcon:SetPoint("LEFT", self.power, "RIGHT", 10, 0)
		end
	end
end

-- Sticky aura colors
local function PostUpdateAuraIcon(self, icons, unit, icon, index, offset, filter, isDebuff)
	-- We want the border, not the color for the type indication
	icon.overlay:SetVertexColor(1, 1, 1)

	local _, _, _, _, dtype, _, _, caster, _, _, _ = UnitAura(unit, index, filter)

	if filter == "HELPFUL" and not UnitCanAttack("player", self.unit) and caster == "player" and cfg["Auras" .. string.upper(self.unit)].StickyAuras.myBuffs then
		-- Sticky aura: myBuffs
		icon.icon:SetVertexColor(unpack(cfg.colors.text))
		icon.icon:SetDesaturated(false)
	elseif filter == "HARMFUL" and UnitCanAttack("player", self.unit) and caster == "player" and cfg["Auras" .. string.upper(self.unit)].StickyAuras.myDebuffs then
		-- Sticky aura: myDebuffs
		icon.icon:SetVertexColor(unpack(cfg.colors.text))
		icon.icon:SetDesaturated(false)
	elseif filter == "HARMFUL" and not UnitCanAttack("player", self.unit) and canDispel[({UnitClass("player")})[2]][dtype] and cfg["Auras" .. string.upper(self.unit)].StickyAuras.curableDebuffs then
		-- Sticky aura: curableDebuffs
		icon.icon:SetVertexColor(DebuffTypeColor[dtype].r, DebuffTypeColor[dtype].g, DebuffTypeColor[dtype].b)
		icon.icon:SetDesaturated(false)
	elseif filter == "HELPFUL" and UnitCanAttack("player", self.unit) and UnitIsUnit(self.unit, caster or "") and cfg["Auras" .. string.upper(self.unit)].StickyAuras.enemySelfBuffs then
		-- Sticky aura: enemySelfBuffs
		icon.icon:SetVertexColor(unpack(cfg.colors.text))
		icon.icon:SetDesaturated(false)
	else
		icon.icon:SetVertexColor(1, 1, 1)
		icon.icon:SetDesaturated(true)
	end
end

-- Custom filters
local customFilter = function(icons, unit, icon, name, rank, texture, count, dtype, duration, timeLeft, caster)
	if icons.filter == "HELPFUL" and not UnitCanAttack("player", unit) and caster == "player" and cfg["Auras" .. string.upper(unit)].StickyAuras.myBuffs then
		-- Sticky aura: myBuffs
		return true
	elseif icons.filter == "HARMFUL" and UnitCanAttack("player", unit) and caster == "player" and cfg["Auras" .. string.upper(unit)].StickyAuras.myDebuffs then
		-- Sticky aura: myDebuffs
		return true
	elseif icons.filter == "HARMFUL" and not UnitCanAttack("player", unit) and canDispel[({UnitClass("player")})[2]][dtype] and cfg["Auras" .. string.upper(unit)].StickyAuras.curableDebuffs then
		-- Sticky aura: curableDebuffs
		return true
	-- Usage of UnitIsUnit: Call from within focus frame will return "target" as caster if focus is targeted (player > target > focus)
	elseif icons.filter == "HELPFUL" and UnitCanAttack("player", unit) and UnitIsUnit(unit, caster or "") and cfg["Auras" .. string.upper(unit)].StickyAuras.enemySelfBuffs then
		-- Sticky aura: enemySelfBuffs
		return true
	else
		-- Aura is not sticky, filter is set to blacklist
		if cfg["Auras" .. string.upper(unit)].FilterMethod[icons.filter == "HELPFUL" and "Buffs" or "Debuffs"] == "BLACKLIST" then
			for _, v in ipairs(cfg["Auras" .. string.upper(unit)].BlackList) do
				if v == name then
					return false
				end
			end
			return true
		-- Aura is not sticky, filter is set to whitelist
		elseif cfg["Auras" .. string.upper(unit)].FilterMethod[icons.filter == "HELPFUL" and "Buffs" or "Debuffs"] == "WHITELIST" then
			for _, v in ipairs(cfg["Auras" .. string.upper(unit)].WhiteList) do
				if v == name then
					return true
				end
			end
			return false
		-- Aura is not sticky, filter is set to none
		else
			return true
		end
	end
end

-- Aura mouseover
local function OnEnterAura(self, isDebuff)
	-- Aura magnification
	if isDebuff then
		self.frame.HighlightAura:SetSize(cfg.DebuffSize * cfg.AuraMagnification, cfg.DebuffSize * cfg.AuraMagnification)
		self.frame.HighlightAura.icon:SetSize(cfg.DebuffSize * cfg.AuraMagnification, cfg.DebuffSize * cfg.AuraMagnification)
		self.frame.HighlightAura:SetPoint("TOPLEFT", self, "TOPLEFT", -(cfg.DebuffSize * cfg.AuraMagnification - cfg.DebuffSize) / 2, (cfg.DebuffSize * cfg.AuraMagnification - cfg.DebuffSize) / 2)
	else
		self.frame.HighlightAura:SetSize(cfg.BuffSize * cfg.AuraMagnification, cfg.BuffSize * cfg.AuraMagnification)
		self.frame.HighlightAura.icon:SetSize(cfg.BuffSize * cfg.AuraMagnification, cfg.BuffSize * cfg.AuraMagnification)
		self.frame.HighlightAura:SetPoint("TOPLEFT", self, "TOPLEFT", -(cfg.BuffSize * cfg.AuraMagnification - cfg.BuffSize) / 2, (cfg.BuffSize * cfg.AuraMagnification - cfg.BuffSize) / 2)
	end
	self.frame.HighlightAura.icon:SetTexture(self.icon:GetTexture())
	self.frame.HighlightAura:Show()

	-- Add aura owner to tooltip if available and not player
	local _, _, _, _, _, _, _, unitCaster, _, _, _ = UnitAura(self.frame.unit, self:GetID(), self.filter)
	if unitCaster and unitCaster ~= "player" then
		local unitCasterName = UnitName(unitCaster)
		if unitCasterName then
			if isDebuff then
				GameTooltip:AddLine("(Debuffed by " .. unitCasterName .. ")")
			else
				GameTooltip:AddLine("(Buffed by " .. unitCasterName .. ")")
			end
		end
	end
	GameTooltip:Show()
end

-- Aura mouseout
local function OnLeaveAura(self, isDebuff)
	self.frame.HighlightAura:Hide()
end

-- Hook aura scripts, set aura border
local function PostCreateAuraIcon(self, icon, icons, index, isDebuff)
	if IsAddOnLoaded("ButtonFacade") then
		-- ButtonFacade support
		BFAuras:AddButton(icon, {
			Icon = icon.icon,
			Cooldown = icon.cd,
			Pushed = false,
			Flash = false,
			Normal = icon.overlay,
			Disabled = false,
			Checked = false,
			Border = false,
			Highlight = false,
			AutoCast = BFDummy,
			AutoCastable = false,
			HotKey = false,
			Count = false,
			Name = false,
		})
		icons.showType = true -- Show border
	elseif cfg.AuraBorder then
		-- Custom aura border
		icon.overlay:SetTexture(cfg.AuraBorder)
		icon.overlay:SetPoint("TOPLEFT", icon, "TOPLEFT", -2, 2)
		icon.overlay:SetPoint("BOTTOMRIGHT", icon, "BOTTOMRIGHT", 2, -2)
		icon.overlay:SetTexCoord(0, 1, 0, 1)
		icons.showType = true
	end

	icon:HookScript("OnEnter", function(self, event, ...) OnEnterAura(self, isDebuff) end)
	icon:HookScript("OnLeave", function(self, event, ...) OnLeaveAura(self, isDebuff) end)
	-- Cancel player buffs on right click
	icon:HookScript("OnClick", function(self, button, down)
		if button == "RightButton" and down == false then
			if self.filter == "HELPFUL" and UnitIsUnit("player", icons:GetParent().unit) then
				CancelUnitBuff("player", self:GetID())
				OnLeaveAura(self, false)
			end
		end
	end)
end

-- Debuff anchoring
local function PreAuraSetPosition(self, auras, max)
	if auras == self.Buffs then
		-- max & visibleBuffs don't work with customFilters, so we have to count
		local num = 0
		for i = 1, self.Buffs.num do
			if not self.Buffs[i] or not self.Buffs[i]:IsVisible() then break end
			num = i
		end
		if num > 0 then
			-- Anchor debuff frame to bottomost buff icon, i.e the last buff row
			self.Debuffs:SetPoint("TOP", auras[num], "BOTTOM", 0, -cfg.AuraSpacing -2)
		else
			-- No buffs
			if self.CPoints then
				self.Debuffs:SetPoint("TOP", self.CPoints[1], "BOTTOM", 0, -10)
			else
				self.Debuffs:SetPoint("TOP", self, "BOTTOM", 0, -10)
			end
		end
	end
end


-- Custom Tags -----------------------------------

-- "163.5k/1.32m"
oUF.TagEvents["[healthDetailed]"] = "UNIT_HEALTH UNIT_MAXHEALTH"
oUF.Tags["[healthDetailed]"] = function(unit)
	return ("%s/%s"):format(valShort(UnitHealth(unit)), valShort(UnitHealthMax(unit)))
end

-- "23.5k/40.9k"
oUF.TagEvents["[ppDetailed]"] = "UNIT_MAXENERGY UNIT_MAXFOCUS UNIT_MAXMANA UNIT_MAXRAGE UNIT_ENERGY UNIT_FOCUS UNIT_MANA UNIT_RAGE UNIT_MAXRUNIC_POWER UNIT_RUNIC_POWER"
oUF.Tags["[ppDetailed]"] = function(unit)
	local _, pType = UnitPowerType(unit)
	local color = cfg.colors.power[pType] or cfg.colors.power["POWER_TYPE_STEAM"]
	return ("|cFF%.2x%.2x%.2x%s/%s|r"):format(color[1] * 255, color[2] * 255, color[3] * 255, valShort(UnitPower(unit)), valShort(UnitPowerMax(unit)))
end

-- "163.5k/1.32m | 23.5k/40.9k"
oUF.TagEvents["[h_ppDetailed]"] = "UNIT_HEALTH UNIT_MAXHEALTH UNIT_MAXENERGY UNIT_MAXFOCUS UNIT_MAXMANA UNIT_MAXRAGE UNIT_ENERGY UNIT_FOCUS UNIT_MANA UNIT_RAGE UNIT_MAXRUNIC_POWER UNIT_RUNIC_POWER"
oUF.Tags["[h_ppDetailed]"] = function(unit)
	if UnitPowerMax(unit) == 0 then
		return oUF.Tags["[healthDetailed]"](unit)
	else
		return ("%s || %s"):format(oUF.Tags["[healthDetailed]"](unit), oUF.Tags["[ppDetailed]"](unit))
	end
end

-- ":)"
oUF.TagEvents["[happiness]"] = "UNIT_HAPPINESS"
oUF.Tags["[happiness]"] = function(unit)
	local happiness, _ = GetPetHappiness()
	return ({":/", ":||", ":)"})[happiness]
end

-- "<Afk>Hankthetank"
oUF.TagEvents["[nameDetailed]"] = "PLAYER_FLAGS_CHANGED UNIT_FACTION"
oUF.Tags["[nameDetailed]"] = function(unit)
	local flags, color
	local name = UnitName(unit)
	if UnitIsPlayer(unit) then
		flags = UnitIsAFK(unit) and "<Afk>" or UnitIsDND(unit) and "<Dnd>"
		if GetRealmName(unit) ~= GetRealmName("player") then
			name = name .. "*"
		end
		local _, class = UnitClass(unit)
		color = class and oUF.colors.class[class]
	else
		if UnitIsEnemy(unit, "player") then
			if UnitIsTapped(unit) and not UnitIsTappedByPlayer(unit) then
				color = oUF.colors.tapped
			else
				color = oUF.colors.reaction[1]
			end
		else
			color = oUF.colors.reaction[UnitReaction(unit, "player") or 5] -- 5: Exception for Party Pets (nil)
		end
	end
	return ("|cFF%.2x%.2x%.2x%s%s|r"):format(color[1] * 255, color[2] * 255, color[3] * 255, flags or "", name)
end

-- "Hankthetank", "You"
oUF.Tags["[smartName]"] = function(unit)
	if UnitIsUnit(unit, "player") then
		if unit == "targettarget" and UnitIsEnemy("target", "player") or unit == "targettargettarget" and UnitIsEnemy("targettarget", "player") then
			local color = oUF.colors.reaction[1]
			return ("|cFF%.2x%.2x%.2xYou|r"):format(color[1] * 255, color[2] * 255, color[3] * 255)
		else
			return "You"
		end
	else
		return UnitName(unit)
	end
end

-- Either pet or player name when using a vehicle
oUF.Tags["[petname]"] = function(unit)
	if UnitHasVehicleUI("player") then
		return UnitName("player")
	elseif UnitName("pet") then
		return UnitName("pet")
	else
		return ""
	end
end

oUF.TagEvents["[threatPerc]"] = "UNIT_THREAT_SITUATION_UPDATE"
oUF.Tags["[threatPerc]"] = function(unit)
	_, _, scaledPercent, _, _ = UnitDetailedThreatSituation("player", unit)
	if scaledPercent then
		if cfg.ColorThreat then
			return (" || |cFF%.2x%.2x%.2x%d%%|r"):format(255 * 0.75 * scaledPercent / 100, 255 * 0.75 - (255 * 0.75 * scaledPercent / 100), 0, scaledPercent)
		else
			return (" || %d%%"):format(scaledPercent)
		end
	else
		return ""
	end
end

-- Frame constructor -----------------------------

local function sharedStyle(self, unit)

	self.menu = menu
	self:SetScript("OnEnter", UnitFrame_OnEnter)
	self:SetScript("OnLeave", UnitFrame_OnLeave)
	self:RegisterForClicks"anyup"
	self:SetAttribute("*type2", "menu")

	self.colors = cfg.colors

	local name, power
	if unit == "player" or unit == "target" or unit == "focus" then
		-- HP%
		self:RegisterEvent("UNIT_HEALTH", function(_, _, ...)
			if unit == ... then
				UpdateHealth(self, ...)
			elseif unit == "player" and UnitHasVehicleUI("player") and ... == "pet" then
				UpdateHealth(self, ...)
			end
		end)

		self:RegisterEvent("UNIT_MAXHEALTH", function(_, _, ...)
			if unit == ... then
				UpdateHealth(self, ...)
			elseif unit == "player" and UnitHasVehicleUI("player") and ... == "pet" then
				UpdateHealth(self, ...)
			end
		end)

		local health = {}
		local healthFill = {}

		for i = 4, 1, -1 do
			health[i] = self:CreateTexture(nil, "ARTWORK")
			health[i]:SetTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\digits.blp")
			health[i]:Hide()
			healthFill[i] = self:CreateTexture(nil, "OVERLAY")
			healthFill[i]:SetTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\digits.blp")
			healthFill[i]:SetVertexColor(unpack(cfg.colors.text))
			healthFill[i]:Hide()
		end

		if unit == "player" then
			health[4]:SetPoint("RIGHT")
			health[3]:SetPoint("RIGHT", health[4], "LEFT")
			health[2]:SetPoint("RIGHT", health[3], "LEFT")
			health[1]:SetPoint("RIGHT", health[2], "LEFT")
		elseif unit == "target" or unit == "focus" then
			health[4]:SetPoint("LEFT")
			health[3]:SetPoint("LEFT", health[4], "RIGHT")
			health[2]:SetPoint("LEFT", health[3], "RIGHT")
			health[1]:SetPoint("LEFT", health[2], "RIGHT")
		end

		healthFill[4]:SetPoint("BOTTOM", health[4])
		healthFill[3]:SetPoint("BOTTOM", health[3])
		healthFill[2]:SetPoint("BOTTOM", health[2])
		healthFill[1]:SetPoint("BOTTOM", health[1])

		self.health = health
		self.healthFill = healthFill

		-- Reanchoring handled in UpdateHealth

		-- Power, threat
		power = self:CreateFontString(nil, "OVERLAY")
		power:SetFontObject("UFFontMedium")
		if unit == "player" then power:SetPoint("BOTTOMRIGHT", health[4], "BOTTOMLEFT")
		elseif unit == "target" or unit == "focus" then power:SetPoint("BOTTOMLEFT", health[4], "BOTTOMRIGHT") end
		power.frequentUpdates = true
		if unit == "player" then self:Tag(power, "[ppDetailed]")
		elseif unit == "target" or unit == "focus" then self:Tag(power, cfg.ShowThreat and "[h_ppDetailed][threatPerc]" or "[h_ppDetailed]") end
		self.power = power
	end

	-- Name
	if unit == "target" or unit == "focus" then
		name = self:CreateFontString(nil, "OVERLAY")
		name:SetFontObject("UFFontBig")
		name:SetPoint("BOTTOMLEFT", power, "TOPLEFT")
		self:Tag(name, "[nameDetailed]")
	elseif unit == "pet" then
		name = self:CreateFontString(nil, "OVERLAY")
		name:SetFontObject("UFFontSmall")
		name:SetPoint("RIGHT")
		self:Tag(name, "[happiness] [petname] @[perhp]%")
	elseif unit == "targettarget" or  unit == "targettargettarget" then
		name = self:CreateFontString(nil, "OVERLAY")
		name:SetFontObject("UFFontSmall")
		name:SetPoint("LEFT")
		name.frequentUpdates = true
		if unit == "targettarget" then self:Tag(name, "›  [smartName] @[perhp]%")
		elseif unit == "targettargettarget" then self:Tag(name, "» [smartName] @[perhp]%") end
	end

	self.name = name

	-- Status icons
	if unit == "player" then
		-- Lead
		self.Leader = self:CreateTexture(nil, "OVERLAY")
		self.Leader:SetSize(24, 24)
		self.Leader:SetTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\statusicons.blp")
		self.Leader:SetTexCoord(2 * 24 / 128, 3 * 24 / 128, 0, 24 / 32)

		-- Rested
		self.Resting = self:CreateTexture(nil, "OVERLAY")
		self.Resting:SetSize(24, 24)
		self.Resting:SetTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\statusicons.blp")
		self.Resting:SetTexCoord(24 / 128, 2 * 24 / 128, 0, 24 / 32)

		-- In combat
		self.Combat = self:CreateTexture(nil, "OVERLAY")
		self.Combat:SetSize(24, 24)
		self.Combat:SetTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\statusicons.blp")
		self.Combat:SetTexCoord(0, 24 / 128, 0, 24 / 32)

		-- Anchoring handled in PostUpdateStatus
	end

	-- Raid targets
	if unit == "player" then
		self.RaidIcon = self:CreateTexture(nil, "OVERLAY")
		self.RaidIcon:SetSize(40, 40)
		self.RaidIcon:SetTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\raidicons.blp")
		self.RaidIcon:SetPoint("RIGHT", self.power, "LEFT", -15, 0)
		self.RaidIcon:SetPoint("TOP", self, "TOP", 0, -5)
	elseif unit == "target" or unit == "focus" then
		self.RaidIcon = self:CreateTexture(nil, "OVERLAY")
		self.RaidIcon:SetSize(40, 40)
		self.RaidIcon:SetTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\raidicons.blp")
		self.RaidIcon:SetPoint("LEFT", self.name, "RIGHT", 10, 0)
		self.RaidIcon:SetPoint("TOP", self, "TOP", 0, -5)
	end

	if (cfg.CBEnable) then
		BuildCastbars(self, unit)
	end

	-- Combo points
	if unit == "target" and (select(2, UnitClass("player")) == "ROGUE" or select(2, UnitClass("player")) == "DRUID") then
		local bg = {}
		local fill = {}
		self.CPoints = {}
		for i = 1, MAX_COMBO_POINTS do
			self.CPoints[i] = CreateFrame("Frame", nil, self)
			self.CPoints[i]:SetSize(16, 16)
			if i > 1 then self.CPoints[i]:SetPoint("LEFT", self.CPoints[i - 1], "RIGHT") end
			bg[i] = self.CPoints[i]:CreateTexture(nil, "ARTWORK")
			bg[i]:SetTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\combo.blp")
			bg[i]:SetTexCoord(0, 16 / 64, 0, 1)
			bg[i]:SetAllPoints(self.CPoints[i])
			fill[i] = self.CPoints[i]:CreateTexture(nil, "OVERLAY")
			fill[i]:SetTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\combo.blp")
			fill[i]:SetTexCoord(0.5, 0.75, 0, 1)
			fill[i]:SetVertexColor(unpack(cfg.colors.power.ENERGY))
			fill[i]:SetAllPoints(self.CPoints[i])
		end
		self.CPoints[1]:SetPoint("TOPLEFT", self, "BOTTOMLEFT")
		self.CPoints.unit = "player"
	end

	-- Auras
	if unit == "target" or unit == "focus" then
		-- Buffs
		self.Buffs = CreateFrame("Frame", unit .. "_Buffs", self) -- ButtonFace needs a name
		if self.CPoints then
			self.Buffs:SetPoint("TOPLEFT", self.CPoints[1], "BOTTOMLEFT", 0, -5)
		else
			self.Buffs:SetPoint("TOPLEFT", self, "BOTTOMLEFT", 0, -5)
		end
		self.Buffs:SetHeight(cfg.BuffSize)
		self.Buffs:SetWidth(225)
		self.Buffs.size = cfg.BuffSize
		self.Buffs.spacing = cfg.AuraSpacing
		self.Buffs.initialAnchor = "LEFT"
		self.Buffs["growth-y"] = "DOWN"
		self.Buffs.num = cfg["Auras" .. string.upper(unit)].MaxBuffs
		self.Buffs.filter = "HELPFUL" -- Explicitly set the filter or the first customFilter call won't work

		-- Debuffs
		self.Debuffs = CreateFrame("Frame", unit .. "_Debuffs", self)
		self.Debuffs:SetPoint("LEFT", self, "LEFT", 0, 0)
		self.Debuffs:SetPoint("TOP", self, "TOP", 0, 0) -- We will reanchor this in PreAuraSetPosition
		self.Debuffs:SetHeight(cfg.DebuffSize)
		self.Debuffs:SetWidth(225)
		self.Debuffs.size = cfg.DebuffSize
		self.Debuffs.spacing = cfg.AuraSpacing
		self.Debuffs.initialAnchor = "LEFT"
		self.Debuffs["growth-y"] = "DOWN"
		self.Debuffs.num = cfg["Auras" .. string.upper(unit)].MaxDebuffs
		self.Debuffs.filter = "HARMFUL"

		-- Buff magnification effect on mouseover
		self.HighlightAura = CreateFrame("Frame", nil, self)
		self.HighlightAura:SetFrameLevel(5) -- Above auras (level 3) and their cooldown overlay (4)
		self.HighlightAura.icon = self.HighlightAura:CreateTexture(nil, "OVERLAY")
		self.HighlightAura.icon:SetPoint("TOPLEFT")

		self.PostUpdateAuraIcon = PostUpdateAuraIcon
		self.PostCreateAuraIcon = PostCreateAuraIcon
		self.PreAuraSetPosition = PreAuraSetPosition

		self.CustomAuraFilter = customFilter
	end

	-- Support for oUF_SpellRange. The built-in oUF range check sucks :/
	if (unit == "target" or unit == "focus") and cfg.RangeFade and IsAddOnLoaded("oUF_SpellRange") then
		self.SpellRange = true
		self.inRangeAlpha = 1
		self.outsideRangeAlpha = cfg.RangeFadeOpacity
	end

	-- Runes
	if unit == "player" and select(2, UnitClass("player")) == "DEATHKNIGHT" then
		self.Runes = CreateFrame("Frame", nil, self)
		self.Runes:SetPoint("TOPRIGHT", self, "BOTTOMRIGHT")
		self.Runes:SetSize(96, 16)
		self.Runes.anchor = "TOPLEFT"
		self.Runes.growth = "RIGHT"
		self.Runes.height = 16
		self.Runes.width = 16

		for i = 1, 6 do
			self.Runes[i] = CreateFrame("StatusBar", nil, self.Runes)
			self.Runes[i]:SetStatusBarTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\blank.blp")

			local backdrop = self.Runes[i]:CreateTexture(nil, "ARTWORK")
			backdrop:SetSize(16, 16)
			backdrop:SetAllPoints()
			backdrop:SetTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\combo.blp")
			backdrop:SetTexCoord(0, 16 / 64, 0, 1)

			-- This is actually the fill layer, but "bg" gets automatically vertex-colored by the runebar module. So let's make use of that!
			self.Runes[i].bg = self.Runes[i]:CreateTexture(nil, "OVERLAY")
			self.Runes[i].bg:SetSize(16, 16)
			self.Runes[i].bg:SetPoint("BOTTOM")
			self.Runes[i].bg:SetTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\combo.blp")
			self.Runes[i].bg:SetTexCoord(0.5, 0.75, 0, 1)

			-- Shine effect
			local shinywheee = CreateFrame("Frame", nil, self.Runes[i])
			shinywheee:SetAllPoints()
			shinywheee:SetAlpha(0)
			shinywheee:Hide()

			local shine = shinywheee:CreateTexture(nil, "OVERLAY")
			shine:SetAllPoints()
			shine:SetPoint("CENTER")
			shine:SetTexture("Interface\\Cooldown\\star4.blp")
			shine:SetBlendMode("ADD")

			local anim = shinywheee:CreateAnimationGroup()
			local alphaIn = anim:CreateAnimation("Alpha")
			alphaIn:SetChange(0.3)
			alphaIn:SetDuration(0.4)
			alphaIn:SetOrder(1)
			local rotateIn = anim:CreateAnimation("Rotation")
			rotateIn:SetDegrees(-90)
			rotateIn:SetDuration(0.4)
			rotateIn:SetOrder(1)
			local scaleIn = anim:CreateAnimation("Scale")
			scaleIn:SetScale(2, 2)
			scaleIn:SetOrigin("CENTER", 0, 0)
			scaleIn:SetDuration(0.4)
			scaleIn:SetOrder(1)
			local alphaOut = anim:CreateAnimation("Alpha")
			alphaOut:SetChange(-0.5)
			alphaOut:SetDuration(0.4)
			alphaOut:SetOrder(2)
			local rotateOut = anim:CreateAnimation("Rotation")
			rotateOut:SetDegrees(-90)
			rotateOut:SetDuration(0.3)
			rotateOut:SetOrder(2)
			local scaleOut = anim:CreateAnimation("Scale")
			scaleOut:SetScale(-2, -2)
			scaleOut:SetOrigin("CENTER", 0, 0)
			scaleOut:SetDuration(0.4)
			scaleOut:SetOrder(2)

			anim:SetScript("OnFinished", function() shinywheee:Hide() end)
			shinywheee:SetScript("OnShow", function() anim:Play() end)

			-- Fill the dots
			self.Runes[i]:SetScript("OnValueChanged", function(self, val)
				-- Resize round overlay texture when hidden statusbar changes
				if val == 10 then
					-- Rune ready: show all 16x16px, play animation
					self.bg:SetSize(16, 16)
					self.bg:SetTexCoord(0.5, 0.75, 0, 1)
					shinywheee:Show()
				else
					-- Dot distance from top & bottom of texture: 4px
					self.bg:SetSize(16, 4 + 8 * val / 10)
					-- Show at least the empty 4 bottom pixels + val% of the 8 pixels of the actual dot = 12px max
					self.bg:SetTexCoord(0.25, 0.5, 12 / 16 - 8 * val / 10 / 16, 1)
				end
			end)
		end
	end

	-- Totems (requires oUF_TotemBar)
	if unit == "player" and select(2, UnitClass("player")) == "SHAMAN" and IsAddOnLoaded("oUF_TotemBar") and cfg.TotemBar then
		self.TotemBar = {}
		self.TotemBar.Destroy = cfg.ClickToDestroy
		self.TotemBar.delay = 0.3
		self.TotemBar.colors = cfg.colors.totems
		for i = 1, 4 do
			self.TotemBar[i] = CreateFrame("StatusBar", nil, self)
			if i > 1 then self.TotemBar[i]:SetPoint("RIGHT", self.TotemBar[i - 1], "LEFT", -3 , 0) end
			self.TotemBar[i]:SetStatusBarTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\blank.blp")
			self.TotemBar[i]:SetSize(23, 20)
			self.TotemBar[i]:SetMinMaxValues(0, 1)
			-- DO NOT WANT! ;p
			self.TotemBar[i].bg = self:CreateTexture(nil)

			local backdrop = self.TotemBar[i]:CreateTexture(nil, "ARTWORK")
			backdrop:SetSize(23, 20)
			backdrop:SetAllPoints()
			backdrop:SetTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\totems.blp")
			backdrop:SetTexCoord(0, 23 / 128, 0, 20 / 32)

			local fill = self.TotemBar[i]:CreateTexture(nil, "OVERLAY")
			fill:SetSize(23, 20)
			fill:SetPoint("BOTTOM")
			fill:SetTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\totems.blp")
			fill:SetTexCoord((1 + 23) / 128, ((23 * 2) + 1) / 128, 0, 20 / 32)

			-- Shine effect
			local glowywheee = CreateFrame("Frame", nil, self.TotemBar[i])
			glowywheee:SetAllPoints()
			glowywheee:SetAlpha(0)
			glowywheee:Hide()

			local glow = glowywheee:CreateTexture(nil, "OVERLAY")
			glow:SetAllPoints()
			glow:SetPoint("CENTER")
			glow:SetTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\totems.blp")
			glow:SetTexCoord((2 + 2 * 23) / 128, ((23 * 3) + 2) / 128, 0, 20 / 32)

			local glowend = self.TotemBar[i]:CreateTexture(nil, "OVERLAY")
			glowend:SetAllPoints()
			glowend:SetTexture("Interface\\AddOns\\oUF_Hank_v3\\textures\\totems.blp")
			glowend:SetTexCoord((2 + 2 * 23) / 128, ((23 * 3) + 2) / 128, 0, 20 / 32)
			glowend:SetAlpha(0.5)
			glowend:Hide()

			local anim = glowywheee:CreateAnimationGroup()
			local alphaIn = anim:CreateAnimation("Alpha")
			alphaIn:SetChange(0.5)
			alphaIn:SetSmoothing("OUT")
			alphaIn:SetDuration(1.5)
			alphaIn:SetOrder(1)

			glowywheee:SetScript("OnShow", function()
				glowend:Hide()
				anim:Play()
			end)

			anim:SetScript("OnFinished", function()
				glowend:Show()
				glowend:SetAlpha(0.5)
			end)

			-- Fill the totems
			self.TotemBar[i]:SetScript("OnValueChanged", function(self, val)
				if val == 0 then
					-- Totem expired
					glow:SetVertexColor(self:GetStatusBarColor())
					glowend:SetVertexColor(self:GetStatusBarColor())
					fill:Hide()
					glowywheee:Show()
				else
					fill:SetSize(23, 4 + 12 * val / 1)
					fill:SetTexCoord((1 + 23) / 128, ((23 * 2) + 1) / 128, 16 / 32 - 12 * val / 32, 20 / 32)
					-- 2DO: This eats performance!!
					fill:SetVertexColor(self:GetStatusBarColor())
					if glowywheee:IsVisible() then
						glowywheee:Hide()
						glowend:Hide()
					end
					if not fill:IsVisible() then fill:Show() end
				end
			end)
		end

		self.TotemBar[1]:SetPoint("TOPRIGHT", self, "BOTTOMRIGHT")
	end

	-- Initial size
	if unit == "player" then
		self:SetAttribute("initial-height", 50)
		self:SetAttribute("initial-width", 175)
	elseif  unit == "target" or unit == "focus" then
		self:SetAttribute("initial-height", 50)
		self:SetAttribute("initial-width", 250)
	elseif unit== "pet" or unit == "targettarget" or unit == "targettargettarget" then
		self:SetAttribute("initial-height", 16)
		self:SetAttribute("initial-width", 125)
	end

end

-- Frame creation --------------------------------

oUF:RegisterStyle("Hankv3", sharedStyle)
oUF:SetActiveStyle("Hankv3")
oUF:Spawn("player", "oUF_player"):SetPoint("RIGHT", UIParent, "CENTER", -cfg.FrameMarginX, -cfg.FrameMarginY)
oUF:Spawn("pet", "oUF_pet"):SetPoint("BOTTOMRIGHT", oUF_player, "TOPRIGHT")
oUF:Spawn("target", "oUF_target"):SetPoint("LEFT", UIParent, "CENTER", cfg.FrameMarginX, -cfg.FrameMarginY)
oUF:Spawn("targettarget", "oUF_ToT"):SetPoint("BOTTOMLEFT", oUF_target, "TOPLEFT")
oUF:Spawn("targettargettarget", "oUF_ToTT"):SetPoint("BOTTOMLEFT", oUF_ToT, "TOPLEFT")
oUF:Spawn("focus", "oUF_focus"):SetPoint("CENTER", UIParent, "CENTER", 0, -cfg.FocusFrameMarginY)
oUF_player:SetScale(cfg.FrameScale)
oUF_pet:SetScale(cfg.FrameScale)
oUF_target:SetScale(cfg.FrameScale)
oUF_ToT:SetScale(cfg.FrameScale)
oUF_ToTT:SetScale(cfg.FrameScale)
oUF_focus:SetScale(cfg.FrameScale * cfg.FocusFrameScale)
if cfg.HideParty then HideParty() end

if cfg.RangeFade and not IsAddOnLoaded("oUF_SpellRange") then
	DEFAULT_CHAT_FRAME:AddMessage("oUF_Hank: Please download and install oUF_SpellRange before enabling range checks!", cfg.colors.text[1], cfg.colors.text[2], cfg.colors.text[3])
elseif cfg.TotemBar and not IsAddOnLoaded("oUF_TotemBar") then
	DEFAULT_CHAT_FRAME:AddMessage("oUF_Hank: Please download and install oUF_TotemBar before enabling the totem bar!", cfg.colors.text[1], cfg.colors.text[2], cfg.colors.text[3])
end


--[[ oUF Event Hooks -----------------------------
Some hacky oUF function hooks to bypass limitations
in handling unit frame events
]]

-- Target Hooks ----------------------------------

callback_OnTarget = oUF_target.PLAYER_TARGET_CHANGED -- @@@@ self:PLAYER_ENTERING_WORLD'OnTargetUpdate'
oUF_target.PLAYER_TARGET_CHANGED = function(self, event)
	callback_OnTarget(self, event)
	UpdateHealth(oUF_target, "target")
	if UnitExists("target") then PostUpdateName("target", oUF_target) end
end

oUF_target:HookScript("OnShow", function(self)
	-- Health update on first target
	UpdateHealth(oUF_target, "target")
	PostUpdateName("target", oUF_target)
end)

oUF_target:RegisterEvent("PLAYER_FLAGS_CHANGED", function(_, _, unit)
	if unit == "target" then PostUpdateName("target", oUF_target) end
end)

-- Focus Hooks -----------------------------------

callback_OnFocus = oUF_focus.PLAYER_FOCUS_CHANGED
oUF_focus.PLAYER_FOCUS_CHANGED = function(self, event)
	callback_OnFocus(self, event)
	UpdateHealth(oUF_focus, "focus")
	if UnitExists("focus") then PostUpdateName("focus", oUF_focus) end
end

oUF_focus:HookScript("OnShow", function(self)
	-- Health update on set focus
	UpdateHealth(oUF_focus, "focus")
	PostUpdateName("focus", oUF_focus)
end)

oUF_focus:RegisterEvent("PLAYER_FLAGS_CHANGED", function(_, _, unit)
	if unit == "focus" then PostUpdateName("focus", oUF_focus) end
end)

-- PostUpdateStatus Hooks ------------------------

oUF_player:HookScript("OnEvent", function(self, event, ...)
	if event == "PLAYER_ENTERING_WORLD" then
		-- Health update on login
		UpdateHealth(oUF_player, "player")
		-- Status icons update on login
		PostUpdateStatus(oUF_player)
	end
end)

callback_OnResting = oUF_player.PLAYER_UPDATE_RESTING
oUF_player.PLAYER_UPDATE_RESTING = function(self, event)
	callback_OnResting(self, event)
	-- Status icons on update resting
	PostUpdateStatus(oUF_player)
end

callback_OnRegenDisabled = oUF_player.PLAYER_REGEN_DISABLED
oUF_player.PLAYER_REGEN_DISABLED = function(self, event)
	callback_OnRegenDisabled(self, event)
	-- Status icons update on combat
	PostUpdateStatus(oUF_player)
end

callback_OnRegenEnabled = oUF_player.PLAYER_REGEN_ENABLED
oUF_player.PLAYER_REGEN_ENABLED = function(self, event)
	callback_OnRegenEnabled(self, event)
	-- Status icons update on combat
	PostUpdateStatus(oUF_player)
end

callback_OnLeader = oUF_player.PARTY_LEADER_CHANGED
oUF_player.PARTY_LEADER_CHANGED = function(self, event)
	callback_OnLeader(self, event)
	-- Status icons update on leader change
	PostUpdateStatus(oUF_player)
end
